local AddonName, GGL = ...
--[[-----------------------------------------------------------------------------
Label Widget
Displays text and optionally an icon.
-------------------------------------------------------------------------------]]
local Type, Version = "GoldGoblinLootGUIWidget-ChallengeLine", 1
local AceGUI = LibStub and LibStub("AceGUI-3.0", true)
if not AceGUI or (AceGUI:GetWidgetVersion(Type) or 0) >= Version then return end

-- Lua APIs
local pairs = pairs
local max  = math.max

-- WoW APIs
local CreateFrame, UIParent = CreateFrame, UIParent

-- GGL
local GetMoneyTextShort = GGL.GetMoneyTextShort
local GetMoneyText = GGL.GetMoneyText

local TOP_ITEMS_LIST_SIZE = GGL.CHALLENGE_TT_MAX_TOP_ITEMS
local L = GGL.Locale

--[[-----------------------------------------------------------------------------
Support functions
-------------------------------------------------------------------------------]]
local INVITE_KEY_LIST = GGL.INVITE_KEY_LIST
local INVITE_KEY_LIST_ATLAS_ICONS = GGL.INVITE_KEY_LIST_ATLAS_ICONS

local GAP_SIZE = 2

local FONT_SIZES = {
    [1] = _G.GameFontHighlightLarge,
    [2] = _G.GameFontHighlight,
    [3] = _G.GameFontHighlightSmall,
}
local GOLD_TEXT, SILVER_TEXT, COPPER_TEXT = "|cffffd70ag|r", "|cffc7c7cfs|r", "|cffeda55fc|r"
local PLAYER_OFFLINE_FORMAT = "|cffff0000".."OFF".."|r-%s"
local PLAYER_AFK_FORMAT = "|cfffff468".."AFK".."|r-%s"

local function SetMoneyText(self, money)
    self:SetText(GetMoneyTextShort(money))
end

local function UpdatePoints(self)
    if self.noUpdate then return end
    local frame = self.frame
    local rank = self.rank
    local height = self.name:GetStringHeight() or 20

    -- Set points
    self.image:SetWidth((height*0.5)+4)
    if self.image.stateShown then
        self.image:SetHeight(self.image:GetWidth())
    else
        self.image:SetHeight((height+4))
    end

    rank:SetWidth( ( rank:GetUnboundedStringWidth() > (height+10) ) and ( rank:GetUnboundedStringWidth() ) or ( height+10 ) )
    rank:SetHeight(height)

    self.name:SetPoint("RIGHT", frame, "RIGHT", -(self.value:GetStringWidth()+2), 0)

    -- Set new frame height
    self.resizing = true
	frame:SetHeight(height + 4 + 1)
	frame.height = height + 4
	self.resizing = nil
end

--[[-----------------------------------------------------------------------------
Methods
-------------------------------------------------------------------------------]]
local methods = {
    ["OnAcquire"] = function(self)
        self.sortValue = nil
        self.noUpdate = nil
        self:SetMoney(0, false, true)
        self:SetFontSize(2)
        self:ResetToolTipPoint()
        self:SetInviteState()
        self:SetSessionResetTable()
        self:SetSession()
        self._player = nil
        wipe(self.mergeList)
    end,

    ["OnRelease"] = function(self)
        self:SetPlayer(nil)
        self:SetCustomSortValue(nil)
        self.frame:Hide()
    end,

    ["OnWidthSet"] = function(self, width)
        UpdatePoints(self)
    end,

    ["SetSessionResetTable"] = function(self, resetTable)
        self.resetTable = resetTable
    end,

    ["SetToolTipPoint"] = function(self, ttPoint, parent)
        self.ttPoint = ttPoint or "BOTTOM_X"
        self.ttParent = parent or self.frame
    end,

    ["ResetToolTipPoint"] = function(self)
        self.ttPoint = "BOTTOM_X"
        self.ttParent = self.frame
    end,

    ["SetSession"] = function(self, session)
        self.session = session
    end,

    ["SetPlayer"] = function(self, player)
        if self.player and self.player == player then return end
        if self.player then
            self.player:UnregisterCallbackEvent("PlayerTemplate_DisplayNameUpdate", self.__onDisplayNameUpdate)
            self.player:UnregisterCallbackEvent("PlayerTemplate_PlayerUpdate", self.__onPlayerUpdate)
            self.player = nil
        end
        if player then
            self:SetName(player.displayName)
            --self:SetFaction(player.factionName)
            player:RegisterCallbackEvent("PlayerTemplate_DisplayNameUpdate", self.__onDisplayNameUpdate)
            player:RegisterCallbackEvent("PlayerTemplate_PlayerUpdate", self.__onPlayerUpdate)
            self._player = player   -- saves the last set player
        end
        self.player = player
        self:SetPlayerState(player)
    end,

    --- Unregisters all events that updates the player
    -- If the update is needed again a new SetPlayer call is needed
    ["UnregisterPlayerUpdates"] = function(self)
        if self.player then
            self.player:UnregisterCallbackEvent("PlayerTemplate_DisplayNameUpdate", self.__onDisplayNameUpdate)
            self.player:UnregisterCallbackEvent("PlayerTemplate_PlayerUpdate", self.__onPlayerUpdate)
        end
    end,

    ["GetPlayer"] = function(self)
        return self.player
    end,

    ["SetRank"] = function(self, rank, sizeUpdate)
        if not rank then return end
        local lastRank = self.rankValue

        if lastRank and lastRank ~= rank and not self.inviteState then
            self.image:Show()
            local top, bottom
            local rankDif
            if lastRank > rank then
                -- up
                top = 0
                bottom = 0.5
                rankDif = lastRank - rank - 1
            elseif lastRank < rank then
                -- down
                top = 0.5
                bottom = 1
                rankDif = rank - lastRank - 1
            end
            rankDif = rankDif > 3 and 3 or rankDif
            self.image:SetTexCoord(
                rankDif*0.25,
                (rankDif*0.25)+0.25,
                top,
                bottom
            )
        elseif not self.inviteState then
            self.image:Hide()
        end

        if sizeUpdate then
            self:SetFontSize(rank >= 3 and 3 or rank)
        end

        self.rankValue = rank
        self.rank:SetText(rank..".")
    end,

    ["GetRank"] = function(self)
        return self.rankValue or 0
    end,

    ["SetPlayerState"] = function(self, playerTemplate)
        if not playerTemplate or not self.player or self.player ~= playerTemplate then return end

        if not playerTemplate.isOnline then
            self:SetName(format(PLAYER_OFFLINE_FORMAT, playerTemplate.displayName))
        elseif playerTemplate.isGameAFK then
            self:SetName(format(PLAYER_AFK_FORMAT, playerTemplate.displayName))
        else
            self:SetName(playerTemplate.displayName)
        end

        self._displayName = playerTemplate.displayName
    end,

    ["SetInviteState"] = function(self, state)
        if state == INVITE_KEY_LIST.INVITED then
            self.inviteState = INVITE_KEY_LIST.INVITED
            self.image:SetAtlas(INVITE_KEY_LIST_ATLAS_ICONS[INVITE_KEY_LIST.INVITED])
            self.image:SetTexCoord(0,1,0,1)
            self.image.stateShown = true
            self.image:Show()
        elseif state == INVITE_KEY_LIST.DECLINED then
            self.inviteState = INVITE_KEY_LIST.DECLINED
            self.image:SetAtlas(INVITE_KEY_LIST_ATLAS_ICONS[INVITE_KEY_LIST.DECLINED])
            self.image:SetTexCoord(0,1,0,1)
            self.image.stateShown = true
            self.image:Show()
        elseif self.inviteState then
            -- when we have a inviteState just refresh rank
            local rank = self.rankValue
            self.inviteState = nil
            self.image:SetTexture(GGL.MEDIA_PATH.."RankArrows")
            self.rankValue = nil
            self.image.stateShown = false
            self:SetRank(rank)
        end
        UpdatePoints(self)
    end,

    ["SetName"] = function(self, name)
        self.name:SetText(name)
    end,

    ["GetName"] = function(self, name)
        return  self._displayName
    end,

    ["SetMoney"] = function(self, copper, add, instantUpdate)
        self.money = add and ((self.money or 0) + copper) or (copper or 0)
        if instantUpdate then
            self.value:SetMoneyText(self.money)
            UpdatePoints(self)
        end
        self:UpdateSortValue()
    end,

    ["GetMoney"] = function(self)
        return self.money or 0
    end,

    ["UpdateSortValue"] = function(self)
        self:SetSortValue(self.customSortValue or self.money)
    end,

    ["SetCustomSortValue"] = function(self, value)
        if value then
            self.customSortValue = value
        else
            self.customSortValue = nil
        end
    end,

    ["SetSortValue"] = function(self, value)
        self.sortValue = value or 0
    end,

    ["GetSortValue"] = function(self)
        return self.sortValue or 0
    end,

    ["UpdateListRank"] = function(self, newRank, listUpdate)
        self.noUpdate = true
            self.value:SetMoneyText(self.money)
            self:SetRank(newRank)
            if newRank % 2 == 0 then
                self.bg:Show()
            else
                self.bg:Hide()
            end
        self.noUpdate = false
        UpdatePoints(self)
    end,

    ["SetFontSize"] = function(self, size)
        size = size or 3
        if self.fontSize ~= size and FONT_SIZES[size] then
            self.fontSize = size
            self:SetFontObjectRaw(FONT_SIZES[size])
            UpdatePoints(self)
        end
    end,

	["SetFontRaw"] = function(self, font, height, flags)
        self.rank:SetFont(font, height, flags)
        self.name:SetFont(font, height, flags)
        self.value:SetFont(font, height, flags)
	end,

	["SetFontObjectRaw"] = function(self, font)
        self.rank:SetFont((font or GameFontHighlightSmall):GetFont()) --GameFontNormal
        self.name:SetFont((font or GameFontHighlightSmall):GetFont())
        self.value:SetFont((font or GameFontHighlightSmall):GetFont())
    end,

    ["BeforePerformeUpdate"] = function(self)
        if not self.session or not self.session.statistics or not self.player then return end
        self:SetMoney(self.session.statistics:GetRawGoldForMergeID(self.player) + self.session.statistics:GetItemValueGoldForMergeID(self.player), false)
    end,

    -- self -> targetLine
    ["MergeInto"] = function(self, targetLine)
        if not targetLine or self == targetLine then return end

        local isValidMerge = false

        -- only lines with existing and different player can be merged
        if self.player and targetLine.player and self.player ~= targetLine.player then
            -- check for already merged player conflicts
            if not self.mergeList[targetLine.player] and not targetLine.mergeList[self.player] then
                isValidMerge = true
            end
        end

        -- merge
        if isValidMerge then
            targetLine:SetMoney(self:GetMoney(), true)
            self.mergeList[targetLine.player] = true
        end

        return isValidMerge
    end,
}

--[[-----------------------------------------------------------------------------
Constructor
-------------------------------------------------------------------------------]]
local function Frame_OnEnter(self)
    -- disable tt for player as its not supportet
    if self.obj.player and self.obj.player.isPlayer then return end
    self.ttFrame = GGL.GUI.Statistics:ShowQuickInfoFrame(self.obj.ttParent, "BOTTOMRIGHT", self.obj.ttParent, "BOTTOMLEFT", -5, 0)
    self.ttFrame:UpdateFromStatistics(self.obj.player, self.obj.session.statistics, true)
end

local function Frame_OnLeave(self)
    if not self.ttFrame then return end
    self.ttFrame.frame:Hide()

    self.ttFrame = nil
end

local function Frame_OnClick(self, button)
    --if not self.obj.itemList or not self.obj.ttParent then return end

    --[[
    local ttFrame = GetToolTipFrame()
    if ToolTipFrameIsKeepOpen() then
        if self.obj.player == ttFrame.playerTemplate then
            ttFrame:KeepOpen(false)
            Frame_OnEnter(self)
        else
            ttFrame:KeepOpen(false)
            Frame_OnEnter(self)
            ttFrame:KeepOpen(true)
        end
    else
        ttFrame:KeepOpen(true)
    end
    ]]--
end

local function Constructor()
    local widget = {}
    widget.type  = Type

    local frame = CreateFrame("BUTTON", nil, UIParent)
    frame.obj = widget
    widget.frame = frame
    frame:SetScript("OnEnter", Frame_OnEnter)
    frame:SetScript("OnLeave", Frame_OnLeave)
    frame:SetScript("OnClick", Frame_OnClick)

    local bg = frame:CreateTexture(nil, "BACKGROUND", nil, 1)
    bg:SetAllPoints(frame)
    bg:SetColorTexture(1,1,1,GGL.LINE_ALPHA_VALUE)
    bg:Hide()
    widget.bg = bg

    local image = frame:CreateTexture(nil, "BACKGROUND", nil, 2)
    image:SetPoint("LEFT", frame, "LEFT", 1, 0)
    image:SetTexture(GGL.MEDIA_PATH.."RankArrows")
    image:Hide()
    widget.image = image

    local rank = frame:CreateFontString(nil, "BACKGROUND", "GameFontHighlightSmall", 2)
    rank:SetPoint("LEFT", image, "RIGHT", 0, 0)
    rank:SetJustifyH("CENTER")
    rank:SetJustifyV("RIGHT")
    rank:SetWordWrap(false)
    rank:SetNonSpaceWrap(false)
    widget.rank = rank

    local name = frame:CreateFontString(nil, "BACKGROUND", "GameFontHighlightSmall", 2)
    name:SetPoint("TOP", frame, "TOP", 0, 0)
    name:SetPoint("LEFT", rank, "RIGHT", GAP_SIZE, 0)
    name:SetPoint("BOTTOMRIGHT", frame, "BOTTOMRIGHT", -1, 0)
    name:SetJustifyH("LEFT")
    name:SetJustifyV("MIDDLE")
    name:SetWordWrap(false)
    widget.name = name

    local value = frame:CreateFontString(nil, "BACKGROUND", "GameFontHighlightSmall", 2)
    value:SetPoint("TOPLEFT", rank, "TOPRIGHT", GAP_SIZE, 0)
    value:SetPoint("BOTTOMRIGHT", frame, "BOTTOMRIGHT", -1, 0)
    value:SetJustifyH("RIGHT")
    value:SetJustifyV("MIDDLE")
    value:SetWordWrap(false)
    value.SetMoneyText = SetMoneyText
    widget.value = value

    -- methodes
    for method, func in pairs(methods) do
        widget[method] = func
    end

    -- cb
    widget.__onDisplayNameUpdate = function(playerTemplate, displayName)
        widget:SetPlayerState(playerTemplate)
    end

    widget.__onPlayerUpdate = function(playerTemplate)
        widget:SetPlayerState(playerTemplate)
    end

    -- init
    widget.fontSize = 3
    widget.mergeList = {}
    widget.qualityCount = {numItems = 0}

    return AceGUI:RegisterAsWidget(widget)
end

AceGUI:RegisterWidgetType(Type, Constructor, Version)
